from azure.identity import DefaultAzureCredential
from azfs.clients.client_interface import ClientInterface
from azure.storage.blob import (
    BlobClient,
    ContainerClient
)
from typing import Union


class AzBlobClient(ClientInterface):

    def _get_file_client(
            self,
            storage_account_url: str,
            file_system: str,
            file_path: str,
            credential: Union[DefaultAzureCredential, str]):
        file_client = BlobClient(
            account_url=storage_account_url,
            container_name=file_system,
            blob_name=file_path,
            credential=credential
        )
        return file_client

    def _get_service_client(self):
        raise NotImplementedError

    def _get_container_client(
            self,
            storage_account_url: str,
            file_system: str,
            credential: Union[DefaultAzureCredential, str]):
        container_client = ContainerClient(
            account_url=storage_account_url,
            container_name=file_system,
            credential=credential)
        return container_client

    def _ls(self, path: str, file_path: str):
        blob_list = \
            [f.name for f in self.get_container_client_from_path(path=path).list_blobs(name_starts_with=file_path)]
        return blob_list

    def _get(self, path: str):
        file_bytes = self.get_file_client_from_path(path=path).download_blob().readall()
        return file_bytes

    def _put(self, path: str, data):
        self.get_file_client_from_path(path=path).upload_blob(
            data=data,
            length=len(data),
            overwrite=True
        )
        return True

    def _info(self, path: str):
        return self.get_file_client_from_path(path=path).get_blob_properties()

    def _rm(self, path: str):
        self.get_file_client_from_path(path=path).delete_blob()
        return True



from typing import Union, TypeVar
from azure.identity import DefaultAzureCredential
import io
import gzip
from azfs.utils import BlobPathDecoder

# define the types
FileClient = TypeVar(
    "FileClient",
    "azure.storage.blob.BlobClient",
    "azure.storage.filedatalake.DataLakeFileClient",
    "azure.storage.queue.QueueClient"
)
FileSystemClient = TypeVar(
    "FileSystemClient",
    "azure.storage.blob.ContainerClient",
    "azure.storage.filedatalake.FileSystemClient"
)


class ClientInterface:
    """
    The class provides Azure Blob and Datalake Container and File Client interface.
    Abstract methods below are implemented in each class
        * _get_file_client
        * _get_service_client
        * _get_container_client
        * _ls
        * _get
        * _put
    """

    def __init__(
            self,
            credential: Union[str, DefaultAzureCredential]):
        self.credential = credential

    def get_file_client_from_path(self, path: str) -> FileClient:
        """
        get file_client from given path
        :param path:
        :return:
        """
        storage_account_url, account_kind, file_system, file_path = BlobPathDecoder(path).get_with_url()
        return self._get_file_client(
            storage_account_url=storage_account_url,
            file_system=file_system,
            file_path=file_path,
            credential=self.credential)

    def _get_file_client(
            self,
            storage_account_url: str,
            file_system: str,
            file_path: str,
            credential: Union[DefaultAzureCredential, str]):
        """
        abstract method to be implemented
        get file_client from given path
        :param storage_account_url:
        :param file_system:
        :param file_path:
        :param credential:
        :return:
        """
        raise NotImplementedError

    def _get_service_client(self):
        """
        abstract method to be implemented
        :return:
        """
        raise NotImplementedError

    def get_container_client_from_path(self, path: str) -> FileSystemClient:
        """
        get container_client from given path
        :param path:
        :return:
        """
        storage_account_url, _, file_system, _ = BlobPathDecoder(path).get_with_url()
        return self._get_container_client(
            storage_account_url=storage_account_url,
            file_system=file_system,
            credential=self.credential)

    def _get_container_client(
            self,
            storage_account_url: str,
            file_system: str,
            credential: Union[DefaultAzureCredential, str]):
        """
        abstract method to be implemented
        :param storage_account_url:
        :param file_system:
        :param credential:
        :return:
        """
        raise NotImplementedError

    def ls(self, path: str, file_path: str):
        return self._ls(path=path, file_path=file_path)

    def _ls(self, path: str, file_path: str):
        """
        abstract method to be implemented
        :param path:
        :param file_path:
        :return:
        """
        raise NotImplementedError

    def get(self, path: str):
        """
        download data from Azure Blob or DataLake.
        :param path:
        :return:
        """
        file_bytes = self._get(path=path)

        # gzip圧縮ファイルは一旦ここで展開
        if path.endswith(".gz"):
            file_bytes = gzip.decompress(file_bytes)

        if type(file_bytes) is bytes:
            file_to_read = io.BytesIO(file_bytes)
        else:
            file_to_read = file_bytes
        return file_to_read

    def _get(self, path: str):
        """
        abstract method to be implemented
        :param path:
        :return:
        """
        raise NotImplementedError

    def put(self, path: str, data):
        return self._put(path=path, data=data)

    def _put(self, path: str, data):
        """
        abstract method to be implemented
        :param path:
        :param data:
        :return:
        """
        raise NotImplementedError

    def info(self, path: str):
        return self._info(path=path)

    def _info(self, path: str):
        """
        abstract method to be implemented
        :param path:
        :return:
        """
        raise NotImplementedError

    def rm(self, path: str):
        return self._rm(path=path)

    def _rm(self, path: str):
        """
        abstract method to be implemented
        :param path:
        :return:
        """
        raise NotImplementedError



import pytest
from azfs.clients.blob_client import AzBlobClient
from azfs.clients.datalake_client import AzDataLakeClient
from azfs.clients.client_interface import ClientInterface
import pandas as pd


class TestClientInterface:
    def test_not_implemented_error(self):
        client_interface = ClientInterface(credential="")
        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/test.csv"
        file_path = "test_caontainer"

        with pytest.raises(NotImplementedError):
            client_interface.get(path=path)

        with pytest.raises(NotImplementedError):
            client_interface.put(path=path, data={})

        with pytest.raises(NotImplementedError):
            client_interface.ls(path=path, file_path=file_path)

        with pytest.raises(NotImplementedError):
            client_interface.rm(path=path)

        with pytest.raises(NotImplementedError):
            client_interface.info(path=path)

        with pytest.raises(NotImplementedError):
            client_interface.get_container_client_from_path(path=path)

        with pytest.raises(NotImplementedError):
            client_interface.get_file_client_from_path(path=path)


class TestReadCsv:

    def test_blob_read_csv(self, mocker, _get_csv, var_azc):
        mocker.patch.object(AzBlobClient, "_get", _get_csv)

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/test.csv"

        # read data from not-exist path
        with var_azc:
            df = pd.read_csv_az(path)
        columns = df.columns
        assert "name" in columns
        assert "age" in columns
        assert len(df.index) == 2

    def test_dfs_read_csv(self, mocker, _get_csv, var_azc):
        mocker.patch.object(AzDataLakeClient, "_get", _get_csv)

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/test.csv"

        # read data from not-exist path
        with var_azc:
            df = pd.read_csv_az(path)
        columns = df.columns
        assert "name" in columns
        assert "age" in columns
        assert len(df.index) == 2


class TestReadJson:

    def test_blob_read_json(self, mocker, _get_json, var_azc, var_json):
        mocker.patch.object(AzBlobClient, "_get", _get_json)

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/test.json"

        data = var_azc.read_json(path)
        assert data == var_json

    def test_dfs_read_json(self, mocker, _get_json, var_azc, var_json):
        mocker.patch.object(AzDataLakeClient, "_get", _get_json)

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/test.json"

        data = var_azc.read_json(path)
        assert data == var_json


class TestToCsv:
    def test_blob_to_csv(self, mocker, _put, var_azc, var_df):
        mocker.patch.object(AzBlobClient, "_put", _put)

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/test.csv"

        with var_azc:
            result = var_df.to_csv_az(path)
        assert result

    def test_dfs_to_csv(self, mocker, _put, var_azc, var_df):
        mocker.patch.object(AzDataLakeClient, "_put", _put)

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/test.csv"

        with var_azc:
            result = var_df.to_csv_az(path)
        assert result


class TestLs:
    def test_blob_ls(self, mocker, _ls, var_azc):
        mocker.patch.object(AzBlobClient, "_ls", _ls)

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/"

        file_list = var_azc.ls(path=path)
        assert len(file_list) == 3
        assert "test1.csv" in file_list
        assert "test2.csv" in file_list
        assert "dir/" in file_list

    def test_dfs_ls(self, mocker, _ls, var_azc):
        mocker.patch.object(AzDataLakeClient, "_ls", _ls)

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/"

        file_list = var_azc.ls(path=path)
        assert len(file_list) == 3
        assert "test1.csv" in file_list
        assert "test2.csv" in file_list
        assert "dir/" in file_list


class TestRm:
    def test_blob_rm(self, mocker, _rm, var_azc):
        mocker.patch.object(AzBlobClient, "_rm", _rm)

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/"

        result = var_azc.rm(path=path)
        assert result

    def test_dfs_rm(self, mocker, _rm, var_azc):
        mocker.patch.object(AzDataLakeClient, "_rm", _rm)

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/"

        result = var_azc.rm(path=path)
        assert result


class TestExists:
    def test_blob_exists(self, mocker, _ls, var_azc):
        mocker.patch.object(AzBlobClient, "_ls", _ls)

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/test1.csv"

        result = var_azc.exists(path=path)
        assert result

        # the file below is not exists
        path = "https://testazfs.blob.core.windows.net/test_caontainer/test3.csv"
        result = var_azc.exists(path=path)
        assert not result

    def test_dfs_exists(self, mocker, _ls, var_azc):
        mocker.patch.object(AzDataLakeClient, "_ls", _ls)

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/test1.csv"

        result = var_azc.exists(path=path)
        assert result

        # the file below is not exists
        path = "https://testazfs.dfs.core.windows.net/test_caontainer/test3.csv"
        result = var_azc.exists(path=path)
        assert not result



import pandas as pd
import json
from azure.identity import DefaultAzureCredential
from azfs.clients import AzfsClient
from azfs.error import AzfsInputError
from typing import Union
from azfs.utils import (
    BlobPathDecoder,
    ls_filter
)
import io


class AzFileClient:
    """
    usage:

    ```
    import azfs
    import pandas as pd

    credential = "[your credential]"
    azc = azfs.AzFileClient(credential=credential)

    path = "your blob file url, starts with https://..."
    with azc:
        df = pd.read_csv_az(path)

    with azc:
        df.to_csv_az(path)

    # ls
    file_list = azc.ls(path)
    ```

    """

    def __init__(
            self,
            credential: Union[str, DefaultAzureCredential, None] = None):
        """

        :param credential: if string, Blob Storage -> Access Keys -> Key
        """
        if credential is None:
            credential = DefaultAzureCredential()
        self.credential = credential

    def __enter__(self):
        """
        with句でのread_csv_azとto_csv_azの関数追加処理
        :return:
        """
        pd.__dict__['read_csv_az'] = self.read_csv
        pd.DataFrame.to_csv_az = self.to_csv(self)
        return self

    def __exit__(self, exec_type, exec_value, traceback):
        """
        with句で追加したread_csv_azとto_csv_azの削除
        :param exec_type:
        :param exec_value:
        :param traceback:
        :return:
        """
        pd.__dict__.pop('read_csv_az')
        pd.DataFrame.to_csv_az = None

    @staticmethod
    def to_csv(az_file_client):
        def inner(self, path, **kwargs):
            df = self if isinstance(self, pd.DataFrame) else None
            return az_file_client.write_csv(path=path, df=df, **kwargs)
        return inner

    def exists(self, path: str) -> bool:
        # 親パスの部分を取得
        parent_path = path.rsplit("/", 1)[0]
        file_name = path.rsplit("/", 1)[1]
        file_list = self.ls(parent_path)
        if file_list:
            if file_name in file_list:
                return True
        return False

    def ls(self, path: str):
        """
        list blob file
        :param path:
        :return:
        """
        _, account_kind, _, file_path = BlobPathDecoder(path).get_with_url()
        file_list = AzfsClient.get(account_kind, credential=self.credential).ls(path=path, file_path=file_path)

        return ls_filter(file_path_list=file_list, file_path=file_path)

    def walk(self, path: str, max_depth=2):
        pass

    def cp(self, src_path: str, dst_path: str, overwrite=False):
        """
        copy the data from `src_path` to `dst_path`
        :param src_path:
        :param dst_path:
        :param overwrite:
        :return:
        """
        if src_path == dst_path:
            raise AzfsInputError("src_path and dst_path must be different")
        if (not overwrite) and self.exists(dst_path):
            raise AzfsInputError(f"{dst_path} is already exists. Please set `overwrite=True`.")
        data = self._get(path=src_path)
        if type(data) is io.BytesIO:
            self._put(path=dst_path, data=data.read())
        elif type(data) is bytes:
            self._put(path=dst_path, data=data)
        return True

    def rm(self, path: str) -> bool:
        """
        delete the file in blob
        :param path:
        :return:
        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        return AzfsClient.get(account_kind, credential=self.credential).rm(path=path)

    def info(self, path: str) -> dict:
        """
        get file properties, such as
        * name
        * creation_time
        * last_modified_time
        * size
        * content_hash(md5)
        :param path:
        :return:
        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        # get info from blob or data-lake storage
        data = AzfsClient.get(account_kind, credential=self.credential).info(path=path)

        # extract below to determine file or directory
        content_settings = data.get("content_settings", {})
        metadata = data.get("metadata", {})

        data_type = ""
        if "hdi_isfolder" in metadata:
            # only data-lake storage has `hdi_isfolder`
            data_type = "directory"
        elif content_settings.get("content_type") is not None:
            # blob and data-lake storage have `content_settings`,
            # and its value of the `content_type` must not be None
            data_type = "file"
        return {
            "name": data.get("name", ""),
            "size": data.get("size", ""),
            "creation_time": data.get("creation_time", ""),
            "last_modified": data.get("last_modified", ""),
            "etag": data.get("etag", ""),
            "content_type": content_settings.get("content_type", ""),
            "type": data_type
        }

    def checksum(self, path: str):
        """
        raise KeyError if info has no etag.
        Blob and DataLake storage have etag.
        :param path:
        :return:
        """
        return self.info(path=path)["etag"]

    def size(self, path):
        """
        Size in bytes of file
        """
        return self.info(path).get("size", None)

    def isdir(self, path):
        """
        Is this entry directory-like?
        """
        try:
            return self.info(path)["type"] == "directory"
        except IOError:
            return False

    def isfile(self, path):
        """
        Is this entry file-like?
        """
        try:
            return self.info(path)["type"] == "file"
        except IOError:
            return False

    def glob(self, path: str):
        pass

    def du(self, path):
        pass

    def _get(self, path: str) -> Union[bytes, str, io.BytesIO]:
        """
        storage accountのタイプによってfile_clientを変更し、データを取得する関数
        特定のファイルを取得する関数
        :param path:
        :return:
        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        return AzfsClient.get(account_kind, credential=self.credential).get(path=path)

    def read_csv(self, path: str, **kwargs) -> pd.DataFrame:
        """
        blobにあるcsvを読み込み、pd.DataFrameとして取得する関数。
        gzip圧縮にも対応。
        :param path:
        :return:
        """
        file_to_read = self._get(path)
        return pd.read_csv(file_to_read, **kwargs)

    def _put(self, path: str, data) -> bool:
        """
        upload data to blob or data_lake storage account
        :param path:
        :param data:
        :return:
        """
        _, account_kind, _, _ = BlobPathDecoder(path).get_with_url()
        return AzfsClient.get(account_kind, credential=self.credential).put(path=path, data=data)

    def write_csv(self, path: str, df: pd.DataFrame, **kwargs) -> bool:
        """
        output pandas dataframe to csv file in Datalake storage.
        Note: Unavailable for large loop processing!
        """
        csv_str = df.to_csv(encoding="utf-8", **kwargs)
        return self._put(path=path, data=csv_str)

    def read_json(self, path: str, **kwargs) -> dict:
        """
        read json file in Datalake storage.
        Note: Unavailable for large loop processing!
        """
        file_bytes = self._get(path)
        if type(file_bytes) is io.BytesIO:
            file_bytes = file_bytes.read()
        return json.loads(file_bytes, **kwargs)

    def write_json(self, path: str, data: dict, **kwargs) -> bool:
        """
        output dict to json file in Datalake storage.
        Note: Unavailable for large loop processing!
        """
        return self._put(path=path, data=json.dumps(data, **kwargs))

    # ===================
    # alias for functions
    # ===================

    get = _get
    download = _get
    put = _put
    upload = _put



from azfs.clients.blob_client import AzBlobClient
from azfs.clients.datalake_client import AzDataLakeClient
from azfs.clients.queue_client import AzQueueClient
from typing import Union


class MetaClient(type):
    """
    A metaclass which have AzBlobClient or AzDataLakeClient in class dictionary.
    if another storage type is added, add new storage type as {"***": Class<Az***Client>}
    """
    def __new__(mcs, name, bases, dictionary):
        cls = type.__new__(mcs, name, bases, dictionary)
        # set Clients
        clients = {
            'dfs': AzDataLakeClient,
            'blob': AzBlobClient,
            'queue': AzQueueClient
        }
        cls.CLIENTS = clients
        return cls


class AbstractClient(metaclass=MetaClient):
    pass


class AzfsClient(AbstractClient):
    """
    Interface of AzBlobClient and AzDataLakeClient.
    Different instances can be obtained as below

    blob_client = AzfsClient.get("blob", "***")
    or
    datalake_client = AzfsClient.get("dfs", "***")

    AzfsClient provide easy way to access functions implemented in AzBlobClient and AzDataLakeClient, as below

    # path is azure storage url
    data = AzfsClient.get("blob", "***").download_data(path)

    """
    CLIENTS = {}

    @classmethod
    def get(cls, account_kind: str, credential) -> Union[AzBlobClient, AzDataLakeClient]:
        """
        get AzBlobClient or AzDataLakeClient depending on account_kind
        :param account_kind: currently blob or dfs
        :param credential: AzureDefaultCredential or string
        :return:
        """
        return cls.CLIENTS[account_kind](credential=credential)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="azfs",
    version="0.1.4",
    author="gsy0911",
    author_email="yoshiki0911@gmail.com",
    description="AzFS is to provide convenient Python read/write functions for Azure Storage Account.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gsy0911/azfs",
    packages=setuptools.find_packages(),
    install_requires=[
        "pandas>=1.0.0",
        "azure-identity>=1.3.1",
        "azure-storage-blob>=12.3.0",
    ],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



from azfs.clients.client_interface import ClientInterface
from azfs.utils import import_optional_dependency
from typing import Union
import functools
from azure.identity import DefaultAzureCredential


# import filedatalake optionally
import_filedatalake = functools.partial(
    import_optional_dependency,
    "azure.storage.filedatalake",
    "To handle Azure Datalake storage, the package is required.",
    "azure-storage-file-datalake")


class AzDataLakeClient(ClientInterface):

    def _get_file_client(
            self,
            storage_account_url: str,
            file_system: str,
            file_path: str,
            credential: Union[DefaultAzureCredential, str]):
        filedatalake = import_filedatalake()
        file_client = filedatalake.DataLakeFileClient(
            storage_account_url,
            file_system,
            file_path,
            credential=credential)
        return file_client

    def _get_service_client(self):
        raise NotImplementedError

    def _get_container_client(
            self,
            storage_account_url: str,
            file_system: str,
            credential: Union[DefaultAzureCredential, str]):
        filedatalake = import_filedatalake()
        file_system = filedatalake.FileSystemClient(
            account_url=storage_account_url,
            file_system_name=file_system,
            credential=credential)
        return file_system

    def _ls(self, path: str, file_path: str):
        file_list = \
            [f.name for f in self.get_container_client_from_path(path=path).get_paths(path=file_path, recursive=True)]
        return file_list

    def _get(self, path: str):
        file_bytes = self.get_file_client_from_path(path).download_file().readall()
        return file_bytes

    def _put(self, path: str, data):
        file_client = self.get_file_client_from_path(path=path)
        _ = file_client.create_file()
        _ = file_client.append_data(data=data, offset=0, length=len(data))
        _ = file_client.flush_data(len(data))
        return True

    def _info(self, path: str):
        return self.get_file_client_from_path(path=path).get_file_properties()

    def _rm(self, path: str):
        self.get_file_client_from_path(path=path).delete_file()
        return True



from azfs.clients.client_interface import ClientInterface
from azfs.utils import import_optional_dependency
from typing import Union
import functools
from azure.identity import DefaultAzureCredential


# import filedatalake optionally
import_queue = functools.partial(
    import_optional_dependency,
    "azure.storage.queue",
    "To handle Azure Queue storage, the package is required.",
    "azure-storage-queue")


class AzQueueClient(ClientInterface):

    def _get_file_client(
            self,
            storage_account_url: str,
            file_system: str,
            file_path: str,
            credential: Union[DefaultAzureCredential, str]):
        # import queue optionally
        _queue = import_queue()
        file_client = _queue.QueueClient(
            storage_account_url,
            file_system,
            file_path,
            credential=credential)
        return file_client

    def _get_service_client(self):
        raise NotImplementedError

    def _get_container_client(
            self,
            storage_account_url: str,
            file_system: str,
            credential: Union[DefaultAzureCredential, str]):
        pass

    def _ls(self, path: str, file_path: str):
        pass

    def _get(self, path: str):
        pass

    def _put(self, path: str, data):
        pass

    def _info(self, path: str):
        pass

    def _rm(self, path: str):
        pass



from typing import Union
import importlib
import re
from azfs.error import (
    AzfsInputError
)


class BlobPathDecoder:
    def __init__(self, path: Union[None, str] = None):
        self.storage_account_name = None
        # blob: blob or data_lake: dfs
        self.account_type = None
        self.container_name = None
        self.blob_name = None

        # ここでpathが入った場合はすぐに取得
        if path is not None:
            self.storage_account_name, self.account_type, self.container_name, self.blob_name = self._decode_path(
                path=path)

    @staticmethod
    def _decode_path_storage_account_name(path: str) -> (str, str, str, str):
        url_pattern = r"https://([a-z0-9]*).(dfs|blob).core.windows.net/(.*)"
        result = re.match(url_pattern, path)
        if result:
            storage_account_name = result.group(1)
            account_type = result.group(2)
            container_name = result.group(3)
            return storage_account_name, account_type, container_name, ""
        raise AzfsInputError(f"not matched with {url_pattern}")

    @staticmethod
    def _decode_path_container_name(path: str) -> (str, str, str, str):
        url_pattern = r"https://([a-z0-9]*).(dfs|blob).core.windows.net/(.*?)/"
        result = re.match(url_pattern, path)
        if result:
            storage_account_name = result.group(1)
            account_type = result.group(2)
            container_name = result.group(3)
            return storage_account_name, account_type, container_name, ""
        raise AzfsInputError(f"not matched with {url_pattern}")

    @staticmethod
    def _decode_path_blob_name(path: str) -> (str, str, str, str):
        url_pattern = r"https://([a-z0-9]*).(dfs|blob).core.windows.net/(.*?)/(.+)"
        result = re.match(url_pattern, path)
        if result:
            storage_account_name = result.group(1)
            account_type = result.group(2)
            container_name = result.group(3)
            blob_name = result.group(4)
            return storage_account_name, account_type, container_name, blob_name
        raise AzfsInputError(f"not matched with {url_pattern}")

    @staticmethod
    def _decode_path_without_url(path: str) -> (str, str, str, str):
        url_pattern = r"(dfs|blob)/([a-z0-9]*)/(.+?)/(.*)"
        result = re.match(url_pattern, path)
        if result:
            storage_account_name = result.group(2)
            account_type = result.group(1)
            container_name = result.group(3)
            blob_name = result.group(4)
            return storage_account_name, account_type, container_name, blob_name
        raise AzfsInputError(f"not matched with {url_pattern}")

    @staticmethod
    def _decode_path(path: str) -> (str, str, str, str):
        """
        decode input [path] such as
        * https://([a-z0-9]*).(dfs|blob).core.windows.net/(.*?)/(.*),
        * ([a-z0-9]*)/(.+?)/(.*)

        dfs: data_lake, blob: blob
        :param path:
        :return:
        """
        function_list = [
            BlobPathDecoder._decode_path_blob_name,
            BlobPathDecoder._decode_path_container_name,
            BlobPathDecoder._decode_path_storage_account_name,
            BlobPathDecoder._decode_path_without_url
        ]
        for func in function_list:
            try:
                storage_account_name, account_type, container_name, blob_name = func(path=path)
            except AzfsInputError:
                continue
            else:
                return storage_account_name, account_type, container_name, blob_name
        raise AzfsInputError("合致するパターンがありません")

    def decode(self, path: str):
        self.storage_account_name, self.account_type, self.container_name, self.blob_name = self._decode_path(path=path)
        return self

    def get(self) -> (str, str, str, str):
        return \
            self.storage_account_name, \
            self.account_type, \
            self.container_name, \
            self.blob_name

    def get_with_url(self) -> (str, str, str, str):
        return \
            f"https://{self.storage_account_name}.{self.account_type}.core.windows.net", \
            self.account_type, \
            self.container_name, \
            self.blob_name

# ================ #
# filter based `/` #
# ================ #


def ls_filter(file_path_list: list, file_path: str):
    filtered_list = []
    filtered_list.extend(_ls_get_file_name(file_path_list=file_path_list, file_path=file_path))
    filtered_list.extend(_ls_get_folder_name(file_path_list=file_path_list, file_path=file_path))
    return filtered_list


def _ls_get_file_name(file_path_list: list, file_path: str):
    """
    特定のフォルダ以下にあるファイル名を取得する。
    :param file_path_list:
    :param file_path:
    :return:
    """
    filtered_file_path_list = []
    if not file_path == "":
        file_path_pattern = rf"({file_path}/)(.*)"
        for fp in file_path_list:
            result = re.match(file_path_pattern, fp)
            if result:
                filtered_file_path_list.append(result.group(2))
    else:
        filtered_file_path_list = file_path_list
    return [f for f in filtered_file_path_list if "/" not in f]


def _ls_get_folder_name(file_path_list: list, file_path: str):
    """
    特定のフォルダ以下にあるフォルダ名を取得する
    :param file_path_list:
    :param file_path:
    :return:
    """
    folders_in_file_path = []
    if not file_path == "":
        file_path_pattern = rf"({file_path}/)(.*?/)(.*)"
        for fp in file_path_list:
            result = re.match(file_path_pattern, fp)
            if result:
                folders_in_file_path.append(result.group(2))
    else:
        file_path_pattern = rf"(.*?/)(.*)"
        for fp in file_path_list:
            result = re.match(file_path_pattern, fp)
            if result:
                folders_in_file_path.append(result.group(1))
    return list(set(folders_in_file_path))

# =============== #
# optional import #
# =============== #


def import_optional_dependency(name: str, extra: str = "", pypi_name: str = ""):
    """
    import package dynamically.
    any package can import anywhere, since the package imported is kept singleton.
    :param name: import package name
    :param extra: message to be shown when raise ImportError
    :param pypi_name: name for pip install
    :return:
    """
    msg = (
        f"Missing optional dependency '{name}'. {extra} "
        f"Use pip or conda to install {pypi_name}."
    )
    try:
        module = importlib.import_module(name)
    except ImportError:
        raise ImportError(msg) from None
    return module
